# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test creating a snapshot
"""

# isort: LOCAL
from stratisd_client_dbus import (
    Manager,
    ObjectManager,
    Pool,
    StratisdErrors,
    filesystems,
    get_object,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class CreateSnapshotTestCase(SimTestCase):
    """
    Test with an empty pool.
    """

    _POOLNAME = "deadpool"
    _FSNAME = "some_fs"
    _SNAPSHOTNAME = "ss_fs"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        ((_, (poolpath, _)), _, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        self._pool_object = get_object(poolpath)

        ((_, fs_objects), return_code, _) = Pool.Methods.CreateFilesystems(
            self._pool_object, {"specs": [self._FSNAME]}
        )

        self.assertEqual(return_code, StratisdErrors.OK)

        self._fs_object_path = fs_objects[0][0]
        self.assertNotEqual(self._fs_object_path, "/")

    def test_create(self):
        """
        Test creating a snapshot and ensure that it works.
        """

        ((is_some, ss_object_path), return_code, _) = Pool.Methods.SnapshotFilesystem(
            self._pool_object,
            {"origin": self._fs_object_path, "snapshot_name": self._SNAPSHOTNAME},
        )

        self.assertTrue(is_some)
        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertNotEqual(ss_object_path, "/")

        result = filesystems().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )
        self.assertEqual(len(list(result)), 2)

    def test_duplicate_snapshot_name(self):
        """
        Test creating a snapshot with duplicate name.
        """

        ((is_some, ss_object_path), return_code, _) = Pool.Methods.SnapshotFilesystem(
            self._pool_object,
            {"origin": self._fs_object_path, "snapshot_name": self._SNAPSHOTNAME},
        )

        self.assertTrue(is_some)
        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertNotEqual(ss_object_path, "/")

        (
            (is_some, ss_object_path_dupe_name),
            return_code,
            _,
        ) = Pool.Methods.SnapshotFilesystem(
            self._pool_object,
            {"origin": self._fs_object_path, "snapshot_name": self._SNAPSHOTNAME},
        )

        self.assertFalse(is_some)
        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertEqual(ss_object_path_dupe_name, "/")

        result = filesystems().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )
        self.assertEqual(len(list(result)), 2)



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test accessing properties of a blockdev.
"""

# isort: STDLIB
from random import choice

# isort: LOCAL
from stratisd_client_dbus import Blockdev, Manager, get_object
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class SimpleTestCase(SimTestCase):
    """
    Set up a pool with some blockdevs so that their properties can be checked.
    """

    _POOLNAME = "deadpool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        proxy = get_object(TOP_OBJECT)
        ((_, (_, self._blockdev_object_paths)), _, _) = Manager.Methods.CreatePool(
            proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )

    def test_optional_props(self):
        """
        Test reading and setting some optional blockdev properties.
        """

        bop = get_object(choice(self._blockdev_object_paths))

        (valid, value) = Blockdev.Properties.HardwareInfo.Get(bop)
        self.assertIsInstance(valid, int)
        self.assertIsInstance(value, str)

        (valid, value) = Blockdev.Properties.UserInfo.Get(bop)
        self.assertIsInstance(valid, int)
        self.assertIsInstance(value, str)

        ((changed, _), return_code, _) = Blockdev.Methods.SetUserInfo(
            bop, {"id": (True, "new_id")}
        )

        self.assertEqual(return_code, 0)
        self.assertTrue(changed)

        (valid, value) = Blockdev.Properties.UserInfo.Get(bop)
        self.assertEqual(valid, True)
        self.assertEqual(value, "new_id")

        ((changed, _), return_code, _) = Blockdev.Methods.SetUserInfo(
            bop, {"id": (False, "")}
        )

        self.assertEqual(return_code, 0)
        self.assertTrue(changed)

        (valid, value) = Blockdev.Properties.UserInfo.Get(bop)
        self.assertEqual(valid, False)
        self.assertIsInstance(value, str)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test accessing properties of a filesystem.
"""

# isort: STDLIB
from os.path import isabs

# isort: LOCAL
from stratisd_client_dbus import Filesystem, Manager, Pool, get_object
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class SetNameTestCase(SimTestCase):
    """
    Set up a pool with a name and one filesystem.
    """

    _POOLNAME = "deadpool"
    _FSNAME = "fs"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        proxy = get_object(TOP_OBJECT)
        ((_, (pool_object_path, _)), _, _) = Manager.Methods.CreatePool(
            proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        pool_object = get_object(pool_object_path)
        ((_, created), _, _) = Pool.Methods.CreateFilesystems(
            pool_object, {"specs": [self._FSNAME]}
        )
        self._filesystem_object_path = created[0][0]

    def test_props(self):
        """
        Test reading some filesystem properties.
        """
        filesystem = get_object(self._filesystem_object_path)
        name = Filesystem.Properties.Name.Get(filesystem)

        self.assertEqual(self._FSNAME, name)

        uuid = Filesystem.Properties.Uuid.Get(filesystem)

        # must be a 32 character string
        self.assertEqual(32, len(uuid))

        created = Filesystem.Properties.Created.Get(filesystem)

        # Should be a UTC rfc3339 string, which should end in Z
        self.assertTrue(created.endswith("Z"))
        # I think this is also always true
        self.assertEqual(len(created), 20)

        devnode = Filesystem.Properties.Devnode.Get(filesystem)
        self.assertTrue(isabs(devnode))



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test accessing properties of a blockdev using FetchProperties interface.
"""

# isort: LOCAL
from stratisd_client_dbus import FetchProperties, Manager, get_object
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase


class FetchPropertiesTestCase(SimTestCase):
    """
    Set up a pool with a name.
    """

    _POOLNAME = "fetchprops"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        proxy = get_object(TOP_OBJECT)
        ((_, (_, self._bd_object_paths)), _, _) = Manager.Methods.CreatePool(
            proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": ["/dev/one", "/dev/two", "/dev/red", "/dev/blue"],
            },
        )

    def test_fetch_size_property(self):
        """
        Test FetchProperties for blockdev property, TotalPhysicalSize
        """
        one_blockdev = get_object(self._bd_object_paths[0])

        (size_success, size) = FetchProperties.Methods.GetProperties(
            one_blockdev, {"properties": ["TotalPhysicalSize"]}
        )["TotalPhysicalSize"]

        self.assertEqual(size_success, True)
        self.assertTrue(size.isnumeric())



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test renaming a filesystem.
"""

# isort: LOCAL
from stratisd_client_dbus import (
    Filesystem,
    Manager,
    ObjectManager,
    Pool,
    StratisdErrors,
    filesystems,
    get_object,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class SetNameTestCase(SimTestCase):
    """
    Set up a pool with a name and one filesystem.
    """

    _POOLNAME = "deadpool"
    _FSNAME = "fs"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        ((_, (pool_object_path, _)), _, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        pool_object = get_object(pool_object_path)
        ((_, created), _, _) = Pool.Methods.CreateFilesystems(
            pool_object, {"specs": [self._FSNAME]}
        )
        self._filesystem_object_path = created[0][0]

    def test_null_mapping(self):
        """
        Test rename to same name.
        """
        filesystem = get_object(self._filesystem_object_path)
        ((is_some, result), return_code, _) = Filesystem.Methods.SetName(
            filesystem, {"name": self._FSNAME}
        )

        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertFalse(is_some)
        self.assertEqual(result, "0" * 32)

    def test_new_name(self):
        """
        Test rename to new name.
        """
        filesystem = get_object(self._filesystem_object_path)
        (result, return_code, _) = Filesystem.Methods.SetName(
            filesystem, {"name": "new"}
        )

        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertTrue(result)

        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        (fs_object_path, _) = next(
            filesystems(props={"Name": "new"}).search(managed_objects)
        )
        self.assertEqual(self._filesystem_object_path, fs_object_path)

        fs_object_path = next(
            filesystems(props={"Name": self._FSNAME}).search(managed_objects), None
        )
        self.assertIsNone(fs_object_path)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test 'CreatePool'.
"""

# isort: LOCAL
from stratisd_client_dbus import (
    Manager,
    ObjectManager,
    StratisdErrors,
    get_object,
    pools,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class Create2TestCase(SimTestCase):
    """
    Test 'create'.
    """

    _POOLNAME = "deadpool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        self._devs = _DEVICE_STRATEGY()
        Manager.Methods.ConfigureSimulator(self._proxy, {"denominator": 8})

    def test_create(self):
        """
        Type of result should always be correct.

        If return_code is OK, then pool must exist.
        """
        ((_, (poolpath, devnodes)), return_code, _) = Manager.Methods.CreatePool(
            self._proxy,
            {"name": self._POOLNAME, "redundancy": (True, 0), "devices": self._devs},
        )

        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        all_pools = list(pools().search(managed_objects))
        result = next(
            pools(props={"Name": self._POOLNAME}).search(managed_objects), None
        )

        if return_code == StratisdErrors.OK:
            self.assertIsNotNone(result)
            (pool, _) = result
            self.assertEqual(pool, poolpath)
            self.assertEqual(len(all_pools), 1)
            self.assertLessEqual(len(devnodes), len(self._devs))
        else:
            self.assertIsNone(result)
            self.assertEqual(len(all_pools), 0)

    def test_create_bad_raid(self):
        """
        Creation should always fail if RAID value is wrong.
        """
        (_, return_code, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 1),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        self.assertEqual(return_code, StratisdErrors.ERROR)


class Create3TestCase(SimTestCase):
    """
    Test 'create' on name collision and different arguments.
    """

    _POOLNAME = "deadpool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )

    def test_create_different_blockdevs(self):
        """
        Create should fail trying to create new pool with same name
        and different blockdevs from previous.
        """
        pools1 = pools().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )

        (_, return_code, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        self.assertEqual(return_code, StratisdErrors.ERROR)

        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        pools2 = list(pools().search(managed_objects))
        pool = next(pools(props={"Name": self._POOLNAME}).search(managed_objects), None)

        self.assertIsNotNone(pool)
        self.assertEqual(
            frozenset(x for (x, y) in pools1), frozenset(x for (x, y) in pools2)
        )


class Create4TestCase(SimTestCase):
    """
    Test 'create' when passing same arguments twice (idempotence).
    """

    _POOLNAME = "idempool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        self._blockdevs = ["/dev/one", "/dev/two", "/dev/red", "/dev/blue"]
        Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": self._blockdevs,
            },
        )

    def test_create_same_blockdevs(self):
        """
        Create should succeed trying to create new pool with same name
        and same blockdevs as previous.
        """
        pools1 = pools().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )

        ((is_some, _), return_code, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": self._blockdevs,
            },
        )
        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertFalse(is_some)

        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        pools2 = list(pools().search(managed_objects))
        pool = next(pools(props={"Name": self._POOLNAME}).search(managed_objects), None)

        self.assertIsNotNone(pool)
        self.assertEqual(
            frozenset(x for (x, y) in pools1), frozenset(x for (x, y) in pools2)
        )



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test accessing properties of a filesystem using FetchProperties interface.
"""

# isort: LOCAL
from stratisd_client_dbus import FetchProperties, Manager, Pool, get_object
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class FetchPropertiesTestCase(SimTestCase):
    """
    Set up a pool with a name and a filesystem.
    """

    _POOLNAME = "fetchprops"
    _FSNAME = "fs"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        proxy = get_object(TOP_OBJECT)
        ((_, (pool_object_path, _)), _, _) = Manager.Methods.CreatePool(
            proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        pool_object = get_object(pool_object_path)
        ((_, created), _, _) = Pool.Methods.CreateFilesystems(
            pool_object, {"specs": [self._FSNAME]}
        )
        self._filesystem_object_path = created[0][0]

    def test_fetch_used_property(self):
        """
        Test FetchProperties for filesystem property, Used
        """
        filesystem = get_object(self._filesystem_object_path)

        (used_success, used) = FetchProperties.Methods.GetProperties(
            filesystem, {"properties": ["Used"]}
        )["Used"]

        self.assertEqual(used_success, True)
        self.assertTrue(used.isnumeric())



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test unique stratis instance.
"""


# isort: STDLIB
import os
import subprocess
import unittest

_STRATISD = os.environ["STRATISD"]


class TestUniqueInstance(unittest.TestCase):
    """
    Test that only one instance of stratisd can be running at any given time.
    """

    def setUp(self):
        """
        Start the original stratisd instance. Register a cleanup function to
        terminate it once started.
        """
        self.command_line = [_STRATISD, "--sim", "--debug"]
        self.first_process = subprocess.Popen(
            self.command_line,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            env=os.environ,
        )

        self.addCleanup(self.first_process.terminate)

    def test_unique_instance(self):
        """
        Verify that a second stratisd instance can not be started.
        """
        process = subprocess.Popen(
            self.command_line,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            env=os.environ,
        )
        (_, stderr) = process.communicate()
        self.assertEqual(process.returncode, 1)
        self.assertNotEqual(stderr, "")



# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test the reporting interface in stratisd.
"""

# isort: STDLIB
import json

# isort: LOCAL
from stratisd_client_dbus import ReportR1, get_object
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class GetReportTestCase(SimTestCase):
    """
    Test that getting valid reports succeeds and that getting invalid reports
    fails.
    """

    _POOLNAME = "reportpool"

    def setUp(self):
        """
        Start stratisd.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)

    def test_valid_report_name(self):
        """
        Test that errored_pool_report returns a valid JSON response.
        """
        (json_str, return_code, _) = ReportR1.Methods.GetReport(
            self._proxy, {"name": "errored_pool_report"}
        )
        self.assertEqual(return_code, 0)
        # Test that JSON is valid - will raise ValueError if not
        json.loads(json_str)

    def test_invalid_report_name(self):
        """
        Test that an invalid report name returns an error.
        """
        (json_str, return_code, _) = ReportR1.Methods.GetReport(
            self._proxy, {"name": "nonexistent_report"}
        )
        self.assertEqual(json_str, "")
        self.assertNotEqual(return_code, 0)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test adding blockdevs to a pool.
"""

# isort: LOCAL
from stratisd_client_dbus import (
    Manager,
    ObjectManager,
    Pool,
    StratisdErrors,
    blockdevs,
    get_object,
    pools,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class AddDataDevsTestCase(SimTestCase):
    """
    Test adding devices to a pool which is initially empty.
    """

    _POOLNAME = "deadpool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        self._data_devices = _DEVICE_STRATEGY()
        ((_, (poolpath, self._blockdev_paths)), _, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": self._data_devices,
            },
        )
        self._pool_object = get_object(poolpath)
        Manager.Methods.ConfigureSimulator(self._proxy, {"denominator": 8})

    def test_empty_devs(self):
        """
        Adding an empty list of devs should fail.
        """
        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        (pool, _) = next(pools(props={"Name": self._POOLNAME}).search(managed_objects))

        blockdevs1 = blockdevs(props={"Pool": pool}).search(managed_objects)
        self.assertEqual(len(list(blockdevs1)), len(self._data_devices))

        ((is_some, _), return_code, _) = Pool.Methods.AddDataDevs(
            self._pool_object, {"devices": []}
        )

        self.assertFalse(is_some)
        self.assertEqual(return_code, StratisdErrors.OK)

        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        blockdevs2 = blockdevs(props={"Pool": pool}).search(managed_objects)
        self.assertEqual(len(list(blockdevs2)), len(self._data_devices))

        blockdevs3 = blockdevs(props={}).search(managed_objects)
        self.assertEqual(len(list(blockdevs3)), len(self._data_devices))

    def test_some_devs(self):
        """
        Adding a non-empty list of devs should increase the number of devs
        in the pool.
        """
        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        (pool, _) = next(pools(props={"Name": self._POOLNAME}).search(managed_objects))

        blockdevs1 = blockdevs(props={"Pool": pool}).search(managed_objects)
        self.assertEqual(len(list(blockdevs1)), len(self._data_devices))

        ((is_some, result), return_code, _) = Pool.Methods.AddDataDevs(
            self._pool_object, {"devices": _DEVICE_STRATEGY()}
        )

        num_devices_added = len(result)
        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})

        if return_code == StratisdErrors.OK:
            self.assertTrue(is_some)
            self.assertGreater(num_devices_added, 0)
        else:
            self.assertFalse(is_some)
            self.assertEqual(num_devices_added, 0)

        blockdev_object_paths = frozenset(result)

        # blockdevs exported on the D-Bus are those added and the existing data devs
        blockdevs2 = list(blockdevs(props={"Pool": pool}).search(managed_objects))
        blockdevs2_object_paths = frozenset([op for (op, _) in blockdevs2])
        self.assertEqual(
            blockdevs2_object_paths,
            blockdev_object_paths.union(frozenset(self._blockdev_paths)),
        )

        # no duplicates in the object paths
        self.assertEqual(len(blockdevs2) - len(self._blockdev_paths), num_devices_added)

        # There are no blockdevs but for those in this pool
        blockdevs3 = blockdevs(props={}).search(managed_objects)
        self.assertEqual(
            len(list(blockdevs3)), num_devices_added + len(self._data_devices)
        )

        # There are no cachedevs belonging to this pool
        blockdevs4 = blockdevs(props={"Pool": pool, "Tier": 1}).search(managed_objects)
        self.assertEqual(list(blockdevs4), [])



# Copyright 2019 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test accessing properties of a pool using FetchProperties interface.
"""

# isort: LOCAL
from stratisd_client_dbus import FetchProperties, FetchPropertiesR1, Manager, get_object
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class FetchPropertiesTestCase(SimTestCase):
    """
    Set up a pool with a name.
    """

    _POOLNAME = "fetchprops"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        proxy = get_object(TOP_OBJECT)
        ((_, (pool_object_path, _)), _, _) = Manager.Methods.CreatePool(
            proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        self._pool_object = get_object(pool_object_path)

    def test_fetch_size_property(self):
        """
        Test FetchProperties for pool property, TotalPhysicalSize
        """
        (size_success, size) = FetchProperties.Methods.GetProperties(
            self._pool_object, {"properties": ["TotalPhysicalSize"]}
        )["TotalPhysicalSize"]

        self.assertEqual(size_success, True)
        self.assertTrue(size.isnumeric())

        (size_success, size) = FetchPropertiesR1.Methods.GetProperties(
            self._pool_object, {"properties": ["TotalPhysicalSize"]}
        )["TotalPhysicalSize"]

        self.assertEqual(size_success, True)
        self.assertTrue(size.isnumeric())

    def test_fetch_used_size_property(self):
        """
        Test FetchProperties for pool property, TotalPhysicalUsed
        """
        (size_success, size) = FetchProperties.Methods.GetProperties(
            self._pool_object, {"properties": ["TotalPhysicalUsed"]}
        )["TotalPhysicalUsed"]

        self.assertEqual(size_success, True)
        self.assertTrue(size.isnumeric())

        (size_success, size) = FetchPropertiesR1.Methods.GetProperties(
            self._pool_object, {"properties": ["TotalPhysicalUsed"]}
        )["TotalPhysicalUsed"]

        self.assertEqual(size_success, True)
        self.assertTrue(size.isnumeric())

    def test_fetch_has_cache_property(self):
        """
        Test FetchProperties.r1 for pool HasCache property
        """
        (has_cache_success, has_cache) = FetchPropertiesR1.Methods.GetProperties(
            self._pool_object, {"properties": ["HasCache"]}
        )["HasCache"]

        self.assertEqual(has_cache_success, True)
        # dbus-python Booleans are actually ints, but they define equality
        # that works for bools
        self.assertIn(has_cache, (True, False))

    def test_fetch_data_crypt_key_description(self):
        """
        Test FetchProperties.r1 for pool KeyDescription
        """
        property_name = "KeyDescription"
        (key_desc_success, key_desc) = FetchPropertiesR1.Methods.GetProperties(
            self._pool_object, {"properties": [property_name]}
        )[property_name]

        self.assertEqual(key_desc_success, True)
        self.assertEqual(key_desc, (False, ""))



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test creating a filesystem in a pool.
"""

# isort: LOCAL
from stratisd_client_dbus import (
    Manager,
    ObjectManager,
    Pool,
    StratisdErrors,
    filesystems,
    get_object,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class CreateFSTestCase(SimTestCase):
    """
    Test with an empty pool.
    """

    _POOLNAME = "deadpool"
    _FSNAME = "fs"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        ((_, (poolpath, _)), _, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        self._pool_object = get_object(poolpath)

    def test_create(self):
        """
        Test calling with no actual volume specification. An empty volume
        list should always succeed, and it should not increase the
        number of volumes.
        """
        ((is_some, result), return_code, _) = Pool.Methods.CreateFilesystems(
            self._pool_object, {"specs": []}
        )

        self.assertFalse(is_some)
        self.assertEqual(len(result), 0)
        self.assertEqual(return_code, StratisdErrors.OK)

        result = filesystems().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )
        self.assertEqual(len(list(result)), 0)


class CreateFSTestCase1(SimTestCase):
    """
    Make a filesystem for the pool.
    """

    _POOLNAME = "deadpool"
    _FSNAME = "thunk"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        ((_, (poolpath, _)), _, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        self._pool_object = get_object(poolpath)
        Pool.Methods.CreateFilesystems(self._pool_object, {"specs": [self._FSNAME]})

    def test_create(self):
        """
        Test calling by specifying a volume name. Because there is already
        a volume with the given name, the creation of the new volume should
        return the volume information as unchanged, and no additional volume
        should be created.
        """
        ((is_some, result), return_code, _) = Pool.Methods.CreateFilesystems(
            self._pool_object, {"specs": [self._FSNAME]}
        )

        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertFalse(is_some)
        self.assertEqual(len(result), 0)

        result = filesystems().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )
        self.assertEqual(len(list(result)), 1)

    def test_create_one(self):
        """
        Test calling by specifying a new and different volume name.
        The new volume will be created.
        """
        new_name = "newname"

        ((is_some, result), return_code, _) = Pool.Methods.CreateFilesystems(
            self._pool_object, {"specs": [new_name]}
        )

        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertTrue(is_some)
        self.assertEqual(len(result), 1)

        (_, fs_name) = result[0]
        self.assertEqual(fs_name, new_name)

        result = filesystems().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )
        self.assertEqual(len(list(result)), 2)

    def test_create_multiple(self):
        """
        Test calling by specifying multiple volume names.  Currently multiple
        volume names are not supported due to possible d-bus timeouts.  When
        multiple volume support is added back - this test should be removed.
        """
        ((is_some, result), return_code, _) = Pool.Methods.CreateFilesystems(
            self._pool_object, {"specs": ["a", "b"]}
        )

        self.assertEqual(return_code, StratisdErrors.ERROR)
        self.assertFalse(is_some)
        self.assertEqual(len(result), 0)

        result = filesystems().search(
            ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        )
        self.assertEqual(len(list(result)), 1)



# Copyright 2020 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test accessing properties of a pool.
"""

# isort: LOCAL
from stratisd_client_dbus import ManagerR1, PoolR1, get_object
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class PropertyTestCase(SimTestCase):
    """
    Set up a pool with at least one device.
    """

    _POOLNAME = "deadpool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        proxy = get_object(TOP_OBJECT)
        ((_, (self._pool_object_path, _)), _, _) = ManagerR1.Methods.CreatePool(
            proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
                "key_desc": (False, ""),
            },
        )

    def test_props(self):
        """
        Test reading some pool properties.
        """
        pool = get_object(self._pool_object_path)
        is_encrypted = PoolR1.Properties.Encrypted.Get(pool)

        self.assertEqual(is_encrypted, False)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test DestroyPool.
"""

# isort: LOCAL
from stratisd_client_dbus import (
    Manager,
    MOBlockDev,
    ObjectManager,
    Pool,
    StratisdErrors,
    blockdevs,
    get_object,
    pools,
)
from stratisd_client_dbus._constants import TOP_OBJECT

from .._misc import SimTestCase, device_name_list

_DEVICE_STRATEGY = device_name_list(1)


class Destroy1TestCase(SimTestCase):
    """
    Test 'destroy' on empty database.

    'destroy' should always succeed on an empty database.
    """

    _POOLNAME = "deadpool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)

    def test_execution(self):
        """
        Destroy should succeed since there is nothing to pass to DestroyPool.
        """
        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        pool = next(pools(props={"Name": self._POOLNAME}).search(managed_objects), None)
        self.assertIsNone(pool)

    def test_bogus_object_path(self):
        """
        Success should occur on a bogus object path.
        """
        (_, return_code, _) = Manager.Methods.DestroyPool(self._proxy, {"pool": "/"})
        self.assertEqual(return_code, StratisdErrors.OK)


class Destroy2TestCase(SimTestCase):
    """
    Test 'destroy' on database which contains the given pool and an unknown
    number of devices.
    """

    _POOLNAME = "deadpool"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        self._devices = _DEVICE_STRATEGY()
        Manager.Methods.CreatePool(
            self._proxy,
            {"name": self._POOLNAME, "redundancy": (True, 0), "devices": self._devices},
        )

    def test_execution(self):
        """
        The pool was just created, so it must always be possible to destroy it.
        """
        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        (pool1, _) = next(pools(props={"Name": self._POOLNAME}).search(managed_objects))
        blockdevs1 = blockdevs(props={"Pool": pool1}).search(managed_objects)
        self.assertEqual(
            frozenset(MOBlockDev(b).Devnode() for (_, b) in blockdevs1),
            frozenset(d for d in self._devices),
        )

        (result, return_code, _) = Manager.Methods.DestroyPool(
            self._proxy, {"pool": pool1}
        )

        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        blockdevs2 = blockdevs(props={"Pool": pool1}).search(managed_objects)
        pool2 = next(
            pools(props={"Name": self._POOLNAME}).search(managed_objects), None
        )

        self.assertEqual(return_code, StratisdErrors.OK)
        self.assertIsNone(pool2)
        self.assertTrue(result)
        self.assertEqual(len(list(blockdevs2)), 0)


class Destroy3TestCase(SimTestCase):
    """
    Test 'destroy' on database which contains the given pool and a volume.
    """

    _POOLNAME = "deadpool"
    _FSNAME = "vol"

    def setUp(self):
        """
        Start the stratisd daemon with the simulator.
        Create a pool and a filesystem.
        """
        super().setUp()
        self._proxy = get_object(TOP_OBJECT)
        ((_, (poolpath, _)), _, _) = Manager.Methods.CreatePool(
            self._proxy,
            {
                "name": self._POOLNAME,
                "redundancy": (True, 0),
                "devices": _DEVICE_STRATEGY(),
            },
        )
        Pool.Methods.CreateFilesystems(get_object(poolpath), {"specs": [self._FSNAME]})

    def test_execution(self):
        """
        This should fail since the pool has a filesystem on it.
        """
        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        (pool, _) = next(pools(props={"Name": self._POOLNAME}).search(managed_objects))

        ((is_some, _), return_code, _) = Manager.Methods.DestroyPool(
            self._proxy, {"pool": pool}
        )
        self.assertEqual(return_code, StratisdErrors.BUSY)
        self.assertFalse(is_some)

        managed_objects = ObjectManager.Methods.GetManagedObjects(self._proxy, {})
        (pool1, _) = next(pools(props={"Name": self._POOLNAME}).search(managed_objects))
        self.assertEqual(pool, pool1)
